@@ -148,16 +148,6 @@ public class PstmtPersistenceTest extends CQLTester
 
             if (numberOfEvictedStatements() > 0)
             {
-                assertEquals("Number of statements in table and in cache don't match", numberOfStatementsInMemory(), numberOfStatementsOnDisk());
-
-                // prepare a more statements to trigger more evictions
-                for (int cnt2 = 1; cnt2 < 10; cnt2++)
-                    prepareStatement("INSERT INTO %s (key, val) VALUES (?, ?) USING TIMESTAMP " + cnt2, clientState);
-
-                // each new prepared statement should have caused an eviction
-                assertEquals("eviction count didn't increase by the expected number", numberOfEvictedStatements(), 10);
-                assertEquals("Number of statements in table and in cache don't match", numberOfStatementsInMemory(), numberOfStatementsOnDisk());
-
                 return;
             }
         }
